package org.jvnet.hifaces.application;

import java.io.IOException;

import javax.faces.component.UIComponent;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class StateManager extends javax.faces.application.StateManager {

	private static final Log logger = LogFactory
			.getLog(StateManager.class);

	public static final String SERIALIZED_VIEW_ATTRIBUTE_NAME = StateManager.class
			.getName()
			+ ".SERIALIZED_VIEW";

	private void ensureSavingStateInServer(FacesContext context) {
		if (isSavingStateInClient(context)) {
			throw new IllegalStateException(
					"This state manager only supports server-side state saving.");
		}
	}

	public StateManager() {
		logger.debug("New JspStateManagerImpl instance created");
	}

	protected Object getComponentStateToSave(FacesContext context) {
		final UIViewRoot viewRoot = context.getViewRoot();
		if (viewRoot.isTransient()) {
			return null;
		}
		removeTransientComponents(viewRoot);
		final Object serializedComponentStates = viewRoot
				.processSaveState(context);
		return serializedComponentStates;
	}

	public void removeTransientComponents(UIComponent component) {
		for (int index = component.getChildCount() - 1; index >= 0; index--) {
			final UIComponent child = (UIComponent) component.getChildren()
					.get(index);
			if (child.isTransient()) {
				component.getChildren().remove(index);
			} else {
				removeTransientComponents(child);
			}
		}
	}

	public UIViewRoot restoreView(FacesContext facesContext, String viewId,
			String renderKitId) {
		ensureSavingStateInServer(facesContext);
		final UIViewRoot uiViewRoot = restoreTreeStructure(facesContext,
				viewId, renderKitId);
		if (uiViewRoot != null) {
			uiViewRoot.setViewId(viewId);
			restoreComponentState(facesContext, uiViewRoot, renderKitId);
			final String restoredViewId = uiViewRoot.getViewId();
			if (restoredViewId == null || !(restoredViewId.equals(viewId))) {
				return null;
			}
			removeSerializedViewFromServletSession(facesContext
					.getExternalContext(), viewId);
		}
		return uiViewRoot;
	}

	protected void restoreComponentState(FacesContext facesContext,
			UIViewRoot uiViewRoot, String renderKitId) {

		final SerializedView serializedView = getSerializedViewFromServletSession(
				facesContext.getExternalContext(), uiViewRoot.getViewId());
		if (serializedView == null) {
			logger.error("No serialized view found in server session!");
			// mark UIViewRoot invalid by resetting view id
			uiViewRoot.setViewId(null);
			return;
		}
		final Object serializedComponentStates = serializedView.getState();
		if (serializedComponentStates == null) {
			logger
					.error("No serialized component state found in server session!");
			return;
		}

		if (uiViewRoot.getRenderKitId() == null) {
			// Just to be sure...
			uiViewRoot.setRenderKitId(renderKitId);
		}

		// uiViewRoot.processRestoreState(facesContext,
		// serializedComponentStates);
	}

	public SerializedView saveSerializedView(FacesContext facesContext)
			throws IllegalStateException {
		final ExternalContext externalContext = facesContext
				.getExternalContext();

		final SerializedView serializedView;
		if (externalContext.getRequestMap().containsKey(
				SERIALIZED_VIEW_ATTRIBUTE_NAME)) {
			serializedView = (SerializedView) externalContext.getRequestMap()
					.get(SERIALIZED_VIEW_ATTRIBUTE_NAME);
		} else {
			// first call to saveSerializedView --> create SerializedView
			final Object treeStruct = getTreeStructureToSave(facesContext);
			Object compStates = getComponentStateToSave(facesContext);
			serializedView = new StateManager.SerializedView(treeStruct,
					compStates);
			externalContext.getRequestMap().put(SERIALIZED_VIEW_ATTRIBUTE_NAME,
					serializedView);
		}

		// save state in server session
		saveSerializedViewInServletSession(externalContext, facesContext
				.getViewRoot().getViewId(), serializedView);
		return null; // return null to signal that saving is done
	}

	protected Object getTreeStructureToSave(FacesContext context) {
		UIViewRoot viewRoot = context.getViewRoot();
		if (viewRoot.isTransient()) {
			return null;
		}
		final Object structure = saveViewRoot(viewRoot);
		return structure;
	}

	private Object saveViewRoot(UIViewRoot viewRoot) {
		// TreeStructureManager tsm = new TreeStructureManager();
		// Object structure = tsm.buildTreeStructureToSave(viewRoot);
		// return structure;
		return viewRoot;
	}

	protected UIViewRoot restoreTreeStructure(FacesContext facesContext,
			String viewId, String renderKitId) {
		ensureSavingStateInServer(facesContext);

		final SerializedView serializedView = getSerializedViewFromServletSession(
				facesContext.getExternalContext(), viewId);
		if (serializedView == null) {
			logger.debug("No serialized view found in server session!");
			return null;
		}

		final Object structure = serializedView.getStructure();
		if (structure == null) {
			logger
					.debug("No tree structure state found in server session, former UIViewRoot must have been transient");
			return null;
		}

		final UIViewRoot uiViewRoot = restoreViewRoot(structure);
		logger.debug("Tree structure restored from server session");

		return uiViewRoot;
	}

	private UIViewRoot restoreViewRoot(Object structure) {

		// final TreeStructureManager tsm = new TreeStructureManager();
		// final UIViewRoot uiViewRoot = tsm
		// .restoreTreeStructure((TreeStructureManager.TreeStructComponent)
		// structure);
		// return uiViewRoot;
		return (UIViewRoot) structure;
	}

	public void writeState(FacesContext context, SerializedView state)
			throws IOException {
		ensureSavingStateInServer(context);
	}

	protected void saveSerializedViewInServletSession(
			ExternalContext externalContext, String viewId,
			SerializedView serializedView) {
		// TODO: What, if user has more than one browser window open on the same
		// page?!
		// only the state of the latest accessed window will be stored at the
		// moment
		final Object sv[] = new Object[] { serializedView.getStructure(),
				serializedView.getState() };
		externalContext.getSessionMap().put(
				getSerializedViewKey(externalContext, viewId), sv);
	}

	protected SerializedView getSerializedViewFromServletSession(
			ExternalContext externalContext, String viewId) {
		final String key = getSerializedViewKey(externalContext, viewId);
		if (externalContext.getSessionMap().get(key) == null) {
			return null; // no state information in session
		}

		Object sv[] = (Object[]) externalContext.getSessionMap().get(key);
		return new SerializedView(sv[0], sv[1]);
	}

	protected void removeSerializedViewFromServletSession(
			ExternalContext externalContext, String viewId) {
		externalContext.getSessionMap().remove(
				getSerializedViewKey(externalContext, viewId));
	}

	protected String getSerializedViewKey(ExternalContext externalContext,
			String viewId) {
		return SERIALIZED_VIEW_ATTRIBUTE_NAME + "-" + viewId;
	}
}
